#!/usr/bin/perl
# Lets use a strict pragma.
use strict;
use Getopt::Long;

my ($opt_file, $opt_help);

GetOptions('h'	=> \$opt_help,
			'file=s'	=> \$opt_file) || &usage();
my $debug = 0;

&usage() if $opt_help;
my $fileName = $opt_file || "squidlog.txt";

#open the squid log
open(SQUIDLOG, $fileName) || die("Cannot open log file: $!");

# A hash of requests is declared with IP address as the key.
my (%requests, %domains, $maxDomainSize, %numAccesses, %allDomains, %data);

########################################################################
#
# %requests - A hash of requests with IP address as key.
# %domains - A hash of domains with IP address as the key. Each value of
# the hash is a array of domains accessed by that particular IP. 
#	This array of domain names may contain duplicates.
# %numAccesses - Hash of number of accesses by IP, domainname.
# $numAccesses{10.3.4.5}{"http://slashdot.org"} = 5 means
# IP 10.3.4.5 has accessed http://slashdot.org 5 times in all.
# %allDomains - Number of times each domain was accessed.
# %data - hash of amount of data downloaded with key IP address.
#
########################################################################

my @temp;
my $refDomHash;
my ($hits, $miss, %hitByResponse);
my $oldTimeStamp = -1; # Invalid timestamp, set to negative to indicate entry into the loop.
my @hitsFiveMins;
my $currentHits;
# Clear the screen, could do it a more elegant way but let it be.
`clear`;
while (<SQUIDLOG>)
{
   chomp;
   my($timeStamp, $timeElapsed, $clientIp, $responseCode, $dataSize, $accessMode, $url, $identity, $neighbor, $mimeType) = split (/\s+/, $_);
 if ($debug) {
     print $_;
     print "Time Stamp = $timeStamp\n";
     print "Time Elapsed = $timeElapsed\n";
     print "Client IP = $clientIp\n";
     print "Response Code =  $responseCode\n";
     print "Data Size =  $dataSize\n";
     print "Access Mode =  $accessMode\n";
     print "Url =  $url \n";
     print "Identity = $identity\n";
     print "Neighbor =  $neighbor\n";
     print "Mime Type =  $mimeType\n"; 
 }
 # Increment the request count for this IP.
 $requests{$clientIp}++;
 $data{$clientIp} = $data{$clientIp} + $dataSize;
 if ($responseCode =~ /^TCP_HIT/) { 
 	$hits++;
 } elsif ($responseCode =~ /^TCP_MISS/) {
 	$miss++;
 }
 if ($responseCode =~ /(.*)\/(.*)/) {
	$hitByResponse{$2}++;
 } else {die("Error in the squid log format: $!");}

 # pull out the domain Name part of the URL.
 if ($url =~ /^([a-zA-Z]+:\/\/)(.*\/)/) {
 # http:// kind of entry.
  my $index = index($2,"/");
  my $domainName = $1 . substr($2,0,$index+1);
  if (length($domainName) > $maxDomainSize) { 
 	 $maxDomainSize = length($domainName);
  }
	($debug) && (print $domainName ."\n");
	$numAccesses{$clientIp}{$domainName}++;
	$allDomains{$domainName}++;
        $refDomHash= $domains{$clientIp};
	# Add this domainName to the array of domains that this IP has accessed.
	push (@$refDomHash, $domainName);
	$domains{$clientIp} = \@$refDomHash;
 } elsif ($url !~ /^(\/)(.*\/)/) {
#   regex corresponds to URLS which start with "/", typically local entries e.g. "/scripts/" kinds.
#   Conveniently ignore them
 	($debug) && (print $url . "\n");
 	my $domainName = $url;
 	($debug) && (print $domainName ."\n");
  	$numAccesses{$clientIp}{$domainName}++;
	$allDomains{$domainName}++;
    $refDomHash = $domains{$clientIp};
	push (@$refDomHash, $domainName);
	$domains{$clientIp} = \@$refDomHash;
  	if (length($domainName) > $maxDomainSize) { 
 		 $maxDomainSize = length($domainName);
  	}
 	($debug) &&  (print "MAX DOMAIN SIZE = $maxDomainSize\n");
 }
 # If this is your entry into the loop, set timestamp to timestamp of first entry.
 if ($oldTimeStamp < 0) { $oldTimeStamp = int($timeStamp);}
 if ((int($timeStamp) - $oldTimeStamp) >= 300) {
 	# Push the sum accumulated thus far in the last 5 minutes into an array.
 	push(@hitsFiveMins, $currentHits);
	# Start afresh for next interval.
	$currentHits = 0;
 	$oldTimeStamp += 300 ; # Set to current time stamp.
 }
 $currentHits++;
}

# The sum accumulated in the last 5 minute interval has not yet been pushed into the array.
push(@hitsFiveMins, $currentHits);

#my $domain;
#foreach $domain (sort(keys(%domains)))
#{
#printf("%-${maxDomainSize}s\t%5d\n", $domain, $domains{$domain});
#}
#print the requests sorted by requester ips.
#print the header.
$~ = "CAPTION1";
write;
format CAPTION1 =
*************************@||||||||||||||||||||||||**************************
"Requests by each IP"
.
$~ = "HEADER1";
write;
format HEADER1 =
	@<<<<<<<<<<<<				@>>>>>>>>>
"IP", "REQUESTS"
.
print "*" x 75 . "\n";
my $ip;
foreach $ip (sort sortIP keys(%requests))
{
$~ = "DATA1";
write;
format DATA1 =
	@<<<<<<<<<<<<<<				@>>>>>>>>>>>
$ip, $requests{$ip}
.
}

print "\n";
$~ = "CAPTION2";
write;
format CAPTION2 =
*************************@||||||||||||||||||||||||||||||||||*************************
"Top 5 users with highest requests"
.
$~ = "HEADER2";
write;
format HEADER2 =
	@<<<<<<<<<<<<				@>>>>>>>>>>>>>>>>>>>>>>>
"IP", "REQUESTS"
.
print "*" x 80 . "\n";

my @topFive = sort sortFunc keys(%requests);
# We only need the top five elements in the above array.
for(my $i=0; $i <= 4; $i=$i+1)
{
	printf("%-12s \t %${maxDomainSize}s\n", $topFive[$i], $requests{$topFive[$i]});
}

########################################################################
# Part A :
# 3. The number of unique domains and most frequently accessed domain 
#		 by top 5 ips.
########################################################################
$~ = "CAPTION3";
write;
format CAPTION3 =
***************@|||||||||||||||||||||||||||||||||||||||||||||||||||||||*********
"Num of unique domains & most frequently accessed domains"
.
$~ = "HEADER3";
write;
format HEADER3 =
@|||||||	@|||||||||||||||||||||||||	@||||||||||||||||||||||||
"IP", "Max Accessed Domain", "Number of Unique domains"
.
print "*" x 80 . "\n";
for(my $i=0; $i <= 4; $i=$i+1)
{
	# Declare a temporary variable which is a reference to the hash of domains by ip --> %domains
	my $refDomHash = $domains{$topFive[$i]};
	my $maxAccesses = 0; # Max number of times this Ip has accessed one of the domains
	my $domain;  
	my $maxAccessedDomain; # Domain that was accessed maximum number of times by this ip.
	my @domArray;
	my %seen = ();
	my $item;
	foreach $item (@$refDomHash) {
    push(@domArray, $item) unless $seen{$item}++;
	}
	undef $item, %seen;
	my $uniqueDomains = scalar(@domArray);
	# Loop over the list of various domains that this IP has accessed.
	foreach $domain (@domArray) {
	   ($debug) &&	(print "$domain\n");
		if ($maxAccesses < $numAccesses{$topFive[$i]}{$domain}) { 
		$maxAccesses =    $numAccesses{$topFive[$i]}{$domain};
		$maxAccessedDomain =  $domain;
		}
	}
	printf("%-12s\t%-30s\t%10d\n", $topFive[$i], $maxAccessedDomain, $uniqueDomains);
	undef @domArray;
}
undef @topFive;
########################################################################
# Part A:
# 4. The amount of data downloaded by each ip, and top 5 ips that downloaded largest amount of data.
########################################################################
print "\n";
$~ = "CAPTION4";
write;
format CAPTION4 =
**********************@|||||||||||||||||||||||||||||||||||||||||||**************
"Amount of data Downloaded by each IP"
.
$~ = "HEADER4";
write;
format HEADER4 =
@||||||||||||		@||||||||||||||||||||||||
"IP", "Data Downloaded in Bytes"
.
print "*" x 80 . "\n";
my @topFive = sort sortData keys(%data);
foreach $ip (sort sortIP keys(%data)) {
#    (scalar(@topFive) < 5) && push (@topFive, $ip);
	printf("%-12s \t %15d\n", $ip, $data{$ip});
}

$~ = "CAPTION42";
write;
format CAPTION42 =
*****************@||||||||||||||||||||||||||||||||||****************************
"Top 5 IPs by Downloaded Data"
.
$~ = "HEADER42";
write;
format HEADER42 =
@||||||||||||	@||||||||||||||||||||||||||
"IP", "Data Downloaded in Bytes"
.
print "*" x 80 . "\n";
my @arraySlice = splice(@topFive, 0, 5);
foreach $ip (@arraySlice) {
	printf("%-12s \t %15d \n", $ip, $data{$ip});
}
undef @arraySlice;
 
########################################################################
# Part A:
# 5. Ratio of cache hits to cache misses.
########################################################################
print "\n";
print "*" x 80 . "\n";
printf("Cache hits to misses = %.20f\n",$hits/$miss);

#######################################################################
# Part A:
# 6. Hit By response Code
#######################################################################
print "\n";
my $str = "Hits By Response Code";
$~ = "CAPTION5";
write;
$~ = "HEADER5";
write;
format HEADER5 =
@|||||||||||||||	@|||||||||||||||||||||
"Response Code", "Number of Hits"
.
print "*" x 80 . "\n";
my $response;
foreach $response (sort keys(%hitByResponse)) {
	if ($response !~ /[0-9]{3}/) {
		printf("Unknown Response\t\t%10d\n", $hitByResponse{$response});
	} else {
		printf("%-15d\t\t%10d\n", $response, $hitByResponse{$response});
	}
}

########################################################################
# Part A:
# 7. Total number of hits per 5 minutes.
########################################################################
print "\n";
$str = "Hits Per Five minutes";
$~ = "CAPTION5";
write;
format CAPTION5 =
*****************@||||||||||||||||||||||||||||||||||***************************
$str
.
$~ = "HEADER6";
write;
format HEADER6 =
@|||||||||||||||	@|||||||||||||||||||||
"Interval", "Number of Hits"
.
print "*" x 80 . "\n";
my ($loopIndex, $hitSum, $hitMax) ;
foreach (@hitsFiveMins) {
	if ($hitMax < $hitsFiveMins[$loopIndex]) {$hitMax = $hitsFiveMins[$loopIndex];}
	$hitSum += $hitsFiveMins[$loopIndex];
	printf("%10d\t\t%9d\n", $loopIndex+1, $hitsFiveMins[$loopIndex++]);
}
print "\n";
print "*" x 80 . "\n";
printf("Average Hits per 5 mins = %10.5f\n", $hitSum/scalar(@hitsFiveMins)); 

print "*" x 80 . "\n";
printf("Maximum Hits per 5 mins = %10d\n", $hitMax);
print "*" x 80 . "\n";
print "Num of Unique Domains = " . scalar(keys(%allDomains)) . "\n";
print "\n";

#forget the ip variable defined above.
undef $ip, $response, $loopIndex, $hitSum, $hitMax;

#close the log.
close(SQUIDLOG);

########################################################################
#
# SUBROUTINES
########################################################################

sub sortFunc {
# sort in descending order.
	$requests{$b} <=> $requests{$a}
}
sub sortIP {
# sort in ascending order.
	my @dottedQuadA = split /\./, $a;
	my @dottedQuadB = split /\./, $b;
	my $i;
	for ($i=0; ($dottedQuadA[$i] == $dottedQuadB[$i]); $i++) {;}
	$dottedQuadA[$i] <=> $dottedQuadB[$i];
}

sub sortData {
# sort in descending order.
	$data{$b} <=> $data{$a}
}


sub usage {
my $helpMessage = " 
Usage: LogParse.pl [-h] [-file Squid log]
Where:
	-h	: Prints the usage statement.
	-file:	The squid log file name [default = ./squidlog.txt]\n";
print $helpMessage;
exit;
}
